
!=============================================================!
! File that contains routines to accompany the PP-BCG library !                                            !
!=============================================================!

!--------------------------------------------------------!
! Routines to perform a 2D MM product for dense matrices !
!--------------------------------------------------------!
SUBROUTINE MM_2D(A, b, Ab, timings, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

!VERSION FOR [X * 2X]  PROCESSOR GRIDS

INTEGER, INTENT(IN)                                     :: N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK1, CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(OUT)   :: Ab
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)             :: timings

!Local variables
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE           :: b2, b2g, Ab2
DOUBLE PRECISION                                        :: t1, t2
INTEGER                                                 :: IERR, CART_GRP, col_rank, row_rank
INTEGER, DIMENSION(MPI_STATUS_SIZE)                     :: STATUS

timings = 0.0D0

CALL MPI_COMM_RANK(row_comm, row_rank, IERR)
CALL MPI_COMM_RANK(col_comm, col_rank, IERR)

ALLOCATE(b2(CHUNK1, rhs), b2g(CHUNK1, rhs), Ab2(CHUNK2, rhs))
b2 = 0.0D0; Ab = 0.0D0; Ab2 = 0.0D0; b2g = 0.0D0

!ZERO PROC IN EACH COL BROADCASTS ITS PART OF ARRAY B
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_BCAST(b, CHUNK2*rhs, MPI_DOUBLE_PRECISION, 0, row_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

!DO LOCAL MATRIX-MATRIX MULTIPLICATION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN;  t1 = MPI_WTIME(); ENDIF
CALL DGEMM("N", "N", CHUNK1, rhs, CHUNK2, 1D0, A, CHUNK1, b, CHUNK2, 0D0, b2, CHUNK1)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(1) = timings(1) + (t2-t1); ENDIF

!DO ROW REDUCTION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_REDUCE(b2, b2g, CHUNK1*rhs, MPI_DOUBLE_PRECISION, MPI_SUM, 0, col_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
IF (RANK .EQ. 0) THEN
   Ab = b2g(1:CHUNK2, :)
   Ab2 = b2g(CHUNK2+1:CHUNK1, :)
   CALL MPI_SEND(Ab2, CHUNK2*rhs, MPI_DOUBLE_PRECISION, 1, 1, COMM, IERR)
ENDIF

IF (RANK .NE. 0) THEN
   IF (col_rank .EQ. 0) THEN
      Ab = b2g(1:CHUNK2, :)
      CALL MPI_SEND(Ab, CHUNK2*rhs, MPI_DOUBLE_PRECISION, RANK/nx, 1, COMM, IERR)
      Ab2 = b2g(CHUNK2+1:CHUNK1, :)
      CALL MPI_SEND(Ab2, CHUNK2*rhs, MPI_DOUBLE_PRECISION, RANK/nx+1, 1, COMM, IERR)
      Ab = 0.0D0 ! By doing this, only procs in first row have nonzero Ab
   ENDIF
   IF (row_rank .EQ. 0) THEN
      CALL MPI_RECV(Ab, CHUNK2*rhs, MPI_DOUBLE_PRECISION, (RANK-MOD(RANK,2))*nx, 1, COMM, STATUS, IERR)
   ENDIF
ENDIF
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

DEALLOCATE(b2, b2g, Ab2)

ENDSUBROUTINE MM_2D

!-

SUBROUTINE MM_2D_SQUARE(A, b, Ab, timings, N, CHUNK, nx, rhs, RANK, COMM, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

!VERSION FOR [X * X]  PROCESSOR GRIDS

INTEGER, INTENT(IN)                                     :: N, CHUNK, nx, rhs, RANK, COMM, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK, CHUNK), INTENT(IN)   :: A
DOUBLE PRECISION, DIMENSION(CHUNK, rhs), INTENT(IN)     :: b
DOUBLE PRECISION, DIMENSION(CHUNK, rhs), INTENT(OUT)    :: Ab
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)             :: timings

!Local variables
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE           :: b2, b2g
DOUBLE PRECISION                                        :: t1, t2
INTEGER                                                 :: IERR, CART_GRP, col_rank, row_rank
INTEGER, DIMENSION(MPI_STATUS_SIZE)                     :: STATUS

timings = 0.0D0

CALL MPI_COMM_RANK(row_comm, row_rank, IERR)
CALL MPI_COMM_RANK(col_comm, col_rank, IERR)

ALLOCATE(b2(CHUNK, rhs), b2g(CHUNK, rhs))
b2 = 0.0D0; Ab = 0.0D0; b2g = 0.0D0

!ZERO PROC IN EACH COL BROADCASTS ITS PART OF ARRAY B
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_BCAST(b, CHUNK*rhs, MPI_DOUBLE_PRECISION, 0, row_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

!DO LOCAL MATRIX-MATRIX MULTIPLICATION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN;  t1 = MPI_WTIME(); ENDIF
CALL DGEMM("N", "N", CHUNK, rhs, CHUNK, 1D0, A, CHUNK, b, CHUNK, 0D0, b2, CHUNK)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(1) = timings(1) + (t2-t1); ENDIF

!DO ROW REDUCTION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_REDUCE(b2, b2g, CHUNK*rhs, MPI_DOUBLE_PRECISION, MPI_SUM, 0, col_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

!DO P2P COMMUNICATION -- FIRST COLUMN TO FIRST ROW
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
IF (RANK .EQ. 0) THEN
   Ab = b2g(:, :)
ENDIF

IF (RANK .NE. 0) THEN
   IF (col_rank .EQ. 0) THEN
      CALL MPI_SEND(b2g, CHUNK*rhs, MPI_DOUBLE_PRECISION, RANK/nx, 1, COMM, IERR)
   ENDIF
   IF (row_rank .EQ. 0) THEN
      CALL MPI_RECV(Ab, CHUNK*rhs, MPI_DOUBLE_PRECISION, RANK*nx, 1, COMM, STATUS, IERR)
   ENDIF
ENDIF
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

DEALLOCATE(b2, b2g)

ENDSUBROUTINE MM_2D_SQUARE

!-

SUBROUTINE MM_2D_SKINNY(A, b, Ab, timings, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

!VERSION FOR ANY nx * ny GRID AS LONG AS nx/ny is an integer (nx > ny)

INTEGER, INTENT(IN)                                     :: N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK1, CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(OUT)   :: Ab
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)             :: timings

!Local variables
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE           :: b2, b2g, Ab2
DOUBLE PRECISION                                        :: t1, t2
INTEGER                                                 :: IERR, CART_GRP, col_rank, row_rank, chunks_per_proc, I
INTEGER, DIMENSION(MPI_STATUS_SIZE)                     :: STATUS

timings = 0.0D0
chunks_per_proc = nx/ny

CALL MPI_COMM_RANK(row_comm, row_rank, IERR)
CALL MPI_COMM_RANK(col_comm, col_rank, IERR)

ALLOCATE(b2(CHUNK1, rhs), b2g(CHUNK1, rhs), Ab2(CHUNK2, rhs))
b2 = 0.0D0; Ab = 0.0D0; Ab2 = 0.0D0; b2g = 0.0D0

!ZERO PROC IN EACH COL BROADCASTS ITS PART OF ARRAY B
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_BCAST(b, CHUNK2*rhs, MPI_DOUBLE_PRECISION, 0, row_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

!DO LOCAL MATRIX-MATRIX MULTIPLICATION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN;  t1 = MPI_WTIME(); ENDIF
CALL DGEMM("N", "N", CHUNK1, rhs, CHUNK2, 1D0, A, CHUNK1, b, CHUNK2, 0D0, b2, CHUNK1)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(1) = timings(1) + (t2-t1); ENDIF

!DO ROW REDUCTION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_REDUCE(b2, b2g, CHUNK1*rhs, MPI_DOUBLE_PRECISION, MPI_SUM, 0, col_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
IF (RANK .EQ. 0) THEN
   Ab(1:CHUNK1,:) = b2g(:,:)
ENDIF

IF (RANK .NE. 0) THEN ! this assumes that RANK 0 is also col_rank 0
   IF (col_rank .EQ. 0) THEN
      CALL MPI_SEND(b2g, CHUNK1*rhs, MPI_DOUBLE_PRECISION, row_rank/chunks_per_proc, 1, COMM, IERR)
      Ab = 0.0D0 ! By doing this, only procs in first row have nonzero Ab
   ENDIF
ENDIF

IF (row_rank .EQ. 0) THEN

   IF (RANK .EQ. 0) THEN
      DO I = 2, chunks_per_proc
         CALL MPI_RECV(b2g, CHUNK1*rhs, MPI_DOUBLE_PRECISION, (col_rank*chunks_per_proc + I - 1)*ny, 1, COMM, STATUS, IERR)
         Ab((I-1)*CHUNK1+1:I*CHUNK1,:) = b2g(:,:)
      ENDDO
   ENDIF

   IF (RANK .NE. 0) THEN
      DO I = 1, chunks_per_proc
         CALL MPI_RECV(b2g, CHUNK1*rhs, MPI_DOUBLE_PRECISION, (col_rank*chunks_per_proc + I - 1)*ny, 1, COMM, STATUS, IERR)
         Ab((I-1)*CHUNK1+1:I*CHUNK1,:) = b2g(:,:)
      ENDDO
   ENDIF

ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

DEALLOCATE(b2, b2g, Ab2)

ENDSUBROUTINE MM_2D_SKINNY

!-

SUBROUTINE MM_2D_FAT(A, b, Ab, timings, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm)

IMPLICIT NONE
INCLUDE 'mpif.h'

!VERSION FOR ANY nx * ny GRID AS LONG AS ny/nx is an integer (nx < ny)

INTEGER, INTENT(IN)                                     :: N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK1, CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(OUT)   :: Ab
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)             :: timings

!Local variables
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE           :: b2, b2g, Ab2
DOUBLE PRECISION                                        :: t1, t2
INTEGER                                                 :: IERR, CART_GRP, col_rank, row_rank, I, chunks_per_proc
INTEGER, DIMENSION(MPI_STATUS_SIZE)                     :: STATUS

timings = 0.0D0
chunks_per_proc = ny/nx

CALL MPI_COMM_RANK(row_comm, row_rank, IERR)
CALL MPI_COMM_RANK(col_comm, col_rank, IERR)

ALLOCATE(b2(CHUNK1, rhs), b2g(CHUNK1, rhs), Ab2(CHUNK2, rhs))
b2 = 0.0D0; Ab = 0.0D0; Ab2 = 0.0D0; b2g = 0.0D0

!ZERO PROC IN EACH COL BROADCASTS ITS PART OF ARRAY B
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_BCAST(b, CHUNK2*rhs, MPI_DOUBLE_PRECISION, 0, row_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

!DO LOCAL MATRIX-MATRIX MULTIPLICATION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN;  t1 = MPI_WTIME(); ENDIF
CALL DGEMM("N", "N", CHUNK1, rhs, CHUNK2, 1D0, A, CHUNK1, b, CHUNK2, 0D0, b2, CHUNK1)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(1) = timings(1) + (t2-t1); ENDIF

!DO ROW REDUCTION
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_REDUCE(b2, b2g, CHUNK1*rhs, MPI_DOUBLE_PRECISION, MPI_SUM, 0, col_comm, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
IF (RANK .EQ. 0) THEN
   Ab = b2g(1:CHUNK2, :)
   DO I = 2, chunks_per_proc
      Ab2 = b2g((I-1)*CHUNK2+1:I*CHUNK2, :)
      CALL MPI_SEND(Ab2, CHUNK2*rhs, MPI_DOUBLE_PRECISION, I-1, 1, COMM, IERR)
   ENDDO
ENDIF

IF (RANK .NE. 0) THEN
   IF (col_rank .EQ. 0) THEN
      DO I = 1, chunks_per_proc
         Ab2 = b2g((I-1)*CHUNK2+1:I*CHUNK2, :)
         CALL MPI_SEND(Ab2, CHUNK2*rhs, MPI_DOUBLE_PRECISION, row_rank*chunks_per_proc + I-1, 1, COMM, IERR)
      ENDDO
      Ab = 0.0D0 ! By doing this, only procs in first row have nonzero Ab
   ENDIF
   IF (row_rank .EQ. 0) THEN
      CALL MPI_RECV(Ab, CHUNK2*rhs, MPI_DOUBLE_PRECISION, (RANK/chunks_per_proc)*ny, 1, COMM, STATUS, IERR)
   ENDIF
ENDIF
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings(2) = timings(2) + (t2-t1); ENDIF

DEALLOCATE(b2, b2g, Ab2)

ENDSUBROUTINE MM_2D_FAT

!-

SUBROUTINE MM_2D_WRAPPER(A, b, Ab, timings2, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm)
! MM wrapper routine that executes the appropriate MM routine based on the
! topology of the 2-D grid of processors

IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER, INTENT(IN)                                     :: N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm
DOUBLE PRECISION, DIMENSION(CHUNK1, CHUNK2), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(IN)    :: b
DOUBLE PRECISION, DIMENSION(CHUNK2, rhs), INTENT(OUT)   :: Ab
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)             :: timings2

DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE             :: timings_mm

ALLOCATE(timings_mm(2))
timings_mm = 0.0D0

IF (nx .LT. ny) THEN
  CALL MM_2D_FAT(A, b, Ab, timings_mm, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm)
ENDIF
IF (ny .LT. nx) THEN
  CALL MM_2D_SKINNY(A, b, Ab, timings_mm, N, CHUNK1, CHUNK2, nx, ny, rhs, RANK, COMM, row_comm, col_comm)
ENDIF
IF (ny .EQ. nx) THEN
  CALL MM_2D_SQUARE(A, b, Ab, timings_mm, N, CHUNK1, nx, rhs, RANK, COMM, row_comm, col_comm)
ENDIF

timings2(:) = timings_mm(:)

ENDSUBROUTINE


!-------------------------------------------------------------
! Routines to perform element-wise and block-wise dot products

SUBROUTINE pddot(CHUNK, x, y, glob, timings2, RANK, COMM)
!Element-wise dot product
IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER, INTENT(IN)                             :: CHUNK, RANK, COMM
DOUBLE PRECISION, DIMENSION(CHUNK), INTENT(IN)  :: x, y
DOUBLE PRECISION, INTENT(OUT)                   :: glob
DOUBLE PRECISION                                :: loc, ddot
INTEGER                                         :: IERR
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)     :: timings2
DOUBLE PRECISION                                :: t1, t2

loc = 0.0D0;    glob = 0.0D0

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
loc = ddot(CHUNK, x, 1, y, 1)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_ALLREDUCE(loc, glob, 1, MPI_DOUBLE_PRECISION, MPI_SUM, COMM, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(2) = t2-t1; ENDIF

ENDSUBROUTINE pddot

!--

SUBROUTINE pseudobdot(CHUNK, cols, x, y, glob, timings2, RANK, COMM)
! Element-wise dot product
IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER, INTENT(IN)                                  :: CHUNK, RANK, cols, COMM
DOUBLE PRECISION, DIMENSION(CHUNK,cols), INTENT(IN)  :: x, y
DOUBLE PRECISION, DIMENSION(cols), INTENT(OUT)       :: glob
DOUBLE PRECISION, DIMENSION(cols)                    :: loc
DOUBLE PRECISION                                     :: ddot
INTEGER                                              :: IERR, I
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)          :: timings2
DOUBLE PRECISION                                     :: t1, t2

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
DO I = 1, cols
   loc(I) = ddot(CHUNK, x(1,I), 1, y(1,I), 1)
ENDDO
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_ALLREDUCE(loc, glob, cols, MPI_DOUBLE_PRECISION, MPI_SUM, COMM, IERR)
CALL MPI_BARRIER(MPI_COMM_WORLD, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(2) = t2-t1; ENDIF

ENDSUBROUTINE pseudobdot

!--

SUBROUTINE pbdot(CHUNK, x, y, col, glob, timings2, RANK, COMM)
! Block-wise dot product
IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER, INTENT(IN)                                  :: CHUNK, RANK, col, COMM
DOUBLE PRECISION, DIMENSION(CHUNK, col), INTENT(IN)  :: x, y
DOUBLE PRECISION, DIMENSION(col, col), INTENT(OUT)   :: glob
DOUBLE PRECISION, DIMENSION(col, col)                :: loc
INTEGER                                              :: IERR
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)          :: timings2
DOUBLE PRECISION                                     :: t1, t2

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL DGEMM("T", "N", col, col, CHUNK, 1.0D0, x, CHUNK, y, CHUNK, 0.0D0, loc, col)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_ALLREDUCE(loc, glob, col*col, MPI_DOUBLE_PRECISION, MPI_SUM, COMM, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(2) = t2-t1; ENDIF

ENDSUBROUTINE pbdot

!--

SUBROUTINE pbdot2(CHUNK, x, y, col1, col2, glob, timings2, RANK, COMM)
! Block-wise dot product -- here col1 might differ from col2
IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER, INTENT(IN)                                   :: CHUNK, RANK, col1, col2, COMM
DOUBLE PRECISION, DIMENSION(CHUNK, col1), INTENT(IN)  :: x
DOUBLE PRECISION, DIMENSION(CHUNK, col2), INTENT(IN)  :: y
DOUBLE PRECISION, DIMENSION(col1, col2), INTENT(OUT)  :: glob
DOUBLE PRECISION, DIMENSION(col1, col2)               :: loc
INTEGER                                               :: IERR
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)           :: timings2
DOUBLE PRECISION                                      :: t1, t2


CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL DGEMM("T", "N", col1, col2, CHUNK, 1.0D0, x, CHUNK, y, CHUNK, 0.0D0, loc, col1)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_ALLREDUCE(loc, glob, col1*col2, MPI_DOUBLE_PRECISION, MPI_SUM, COMM, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

ENDSUBROUTINE pbdot2

!

SUBROUTINE pbSdot2(CHUNK, x, y, col1, col2, glob, timings2, RANK, COMM)
! Block-wise dot product -- here col1 might differ from col2
IMPLICIT NONE
INCLUDE 'mpif.h'

INTEGER, INTENT(IN)                          :: CHUNK, RANK, col1, col2, COMM
REAL(4), DIMENSION(CHUNK, col1), INTENT(IN)  :: x
REAL(4), DIMENSION(CHUNK, col2), INTENT(IN)  :: y
REAL(4), DIMENSION(col1, col2), INTENT(OUT)  :: glob
REAL(4), DIMENSION(col1, col2)               :: loc
INTEGER                                      :: IERR
DOUBLE PRECISION, DIMENSION(2), INTENT(OUT)  :: timings2
DOUBLE PRECISION                             :: t1, t2

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL SGEMM("T", "N", col1, col2, CHUNK, 1.0, x, CHUNK, y, CHUNK, 0.0, loc, col1)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t1 = MPI_WTIME(); ENDIF
CALL MPI_ALLREDUCE(loc, glob, col1*col2, MPI_REAL, MPI_SUM, COMM, IERR)
CALL MPI_BARRIER(COMM, IERR)
IF (RANK == 0) THEN; t2 = MPI_WTIME(); timings2(1) = t2-t1; ENDIF

ENDSUBROUTINE pbSdot2


!------------------------------
! Some general purpose routines

SUBROUTINE PSEUDOINVERSE(M, size)
! Computes the pseudoinverse of M (M is of size "size")

IMPLICIT NONE

INTEGER, INTENT(IN)                                   :: size
DOUBLE PRECISION, DIMENSION(size,size), INTENT(INOUT) :: M

DOUBLE PRECISION, DIMENSION(size,size)                :: U, VT, Z, t1, pseud, S
INTEGER                                               :: LDWORK, K, INFO
PARAMETER                                               (LDWORK=10000)
DOUBLE PRECISION, DIMENSION(:),ALLOCATABLE            :: WORK

ALLOCATE(WORK(LDWORK))

CALL DGESVD('A', 'A', size, size, M, size, S, U, size, VT, size, WORK, LDWORK, INFO)

Z = 0.0D0
DO K = 1, size
   IF (S(K, 1) .NE. 0) THEN
      Z(K, K) = 1.0D0 / S(K, 1)
   ENDIF
ENDDO

CALL DGEMM('T', 'N', size, size, size, 1.0D0, VT, size, Z, size, 0.0D0, t1, size)
CALL DGEMM('N', 'T', size, size, size, 1.0D0, t1, size, U, size, 0.0D0, pseud, size)

M = pseud

DEALLOCATE(WORK)

ENDSUBROUTINE PSEUDOINVERSE

!---
SUBROUTINE SYM_BND_EIG(T, N, KD, EIGVLS, EIGVECS)
! Solves a banded, symmetric, eigenvalue problem
IMPLICIT NONE

INTEGER, INTENT(IN)                             :: N, KD
DOUBLE PRECISION, INTENT(IN)   , DIMENSION(N,N) :: T
DOUBLE PRECISION, INTENT(INOUT), DIMENSION(N)   :: EIGVLS
DOUBLE PRECISION, INTENT(INOUT), DIMENSION(N,N) :: EIGVECS

DOUBLE PRECISION, DIMENSION(N,N)    :: Q, Z
DOUBLE PRECISION, DIMENSION(KD+1,N) :: AB
DOUBLE PRECISION, DIMENSION(N)      :: D
DOUBLE PRECISION, DIMENSION(N-1)    :: E
DOUBLE PRECISION,  DIMENSION(N)     :: WORK
INTEGER                             :: I, J, LDAB, INFO, counter, LDWORK, LIWORK
DOUBLE PRECISION, ALLOCATABLE, DIMENSION(:) :: WORK2
INTEGER, ALLOCATABLE, DIMENSION(:)          :: IWORK

LDWORK = 2D0*(1D0 + 4D0*N + N**2)
LIWORK = 6D0*N
LDAB = KD+1
counter = 0

ALLOCATE(WORK2(LDWORK), IWORK(LIWORK))

DO J = 1, N
   DO I = J, J+KD+1
      counter = counter + 1
      AB(counter,J) = T(I,J)
   ENDDO
   counter = 0
ENDDO

!REDUCE TO TRIDIAGONAL FORM
CALL DSBTRD('V', 'L', N, KD, AB, LDAB, D, E, Q, N, WORK, INFO)

!SOLVE TRIDIAGONAL EIGENVALUE PROBLEM
CALL DSTEVD('V', N, D, E, Z, N, WORK2, LDWORK, IWORK, LIWORK, INFO)

!COPY D TO EIGVLS
CALL DCOPY(N, D, 1, EIGVLS, 1)

!FORM Q*Z
CALL DGEMM("N", "N", N, N, N, 1.0D0, Q, N, Z, N, 0.0D0, EIGVECS, N)

DEALLOCATE(WORK2, IWORK)

ENDSUBROUTINE SYM_BND_EIG

